/**
Network server to do hash-based password recovery
homework.

Orion Sky Lawlor, olawlor@acm.org, 2006/2/27 (Public Domain)
*/
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include "socket.h"
#include "socket.cpp" /* include body for easy linking */
#include "sha1.h"
#include "sha1.cpp" /* include body for easy linking */

skt_ip_t client_ip; unsigned int client_port;
char client_name[120];
std::string user;

SOCKET s; /* Connection to client */
void security(const char *why) {
	if (s!=0) {
		skt_sendN(s,why,strlen(why));
		skt_close(s);
	}
	printf("SECURITY ERROR: %s:%d user '%s': %s\n",
			client_name,client_port,user.c_str(),why);
	exit(1);
}

const static struct {
	const char *hint;
} probInfo[]={{"INVALID PROBLEM 0"},
{"Problem 1: I'm thinking of a one-character printable ASCII password"},
{"Problem 2: I'm thinking of a four-digit numeric password"},
{"Problem 3: I'm thinking of a three-character printable ASCII password"},
{"Problem 4: I'm thinking of an eight-digit numeric password, but the digits are in *strictly* increasing order"},
{"Problem 5: I'm thinking of a 10-character password consisting entirely of the lowercase characters 'f' and 'o'"}
};

std::string generate_password(const std::string &user,int probNo)
{
#ifndef SECRET_BIAS /* Secret string used to prevent students from reverse-engineering passwords */
#define SECRET_BIAS "XXX" /* (not the real secret string) */
#endif
	osl::SHA1_hash_t uh=osl::SHA1_hash((user+SECRET_BIAS).c_str());
	/* Seeded random number generator == crappy hash function */
	int seed=SECRET_BIAS[0]+((long *)&uh)[0]+((long *)&uh)[1];
	srand(seed);
	int r=rand();
	printf("User '%s' seed: %d.  User's first random: %08x\n",user.c_str(),seed,r);
	char ret[100];
	memset(ret,0,100);
	char *printableASCII=
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"abcdefghijklmnopqrstuvwxyz"
		"0123456789~!@#$%^&*()_+`-="
		"[]\\;',./{}|:<>?\"";
	int lenASCII=strlen(printableASCII);
	if (probNo==1) {
		ret[0]=printableASCII[rand()%lenASCII];
		return ret;
	}
	if (probNo==2) {
		for (int digit=0;digit<4;digit++)
			ret[digit]='0'+(rand()%10);
		return ret;
	}
	if (probNo==3) {
		for (int digit=0;digit<3;digit++)
			ret[digit]=printableASCII[rand()%lenASCII];
		return ret;
	}
	if (probNo==4) {
		int up1=0+(rand()%8);
		int up2=4+(rand()%8);
		for (int digit=0;digit<8;digit++)
			ret[digit]='0'+digit+(digit>=up1)+(digit>=up2);
		return ret;
	}
	if (probNo==5) {
		for (int digit=0;digit<10;digit++)
			ret[digit]=(rand()%2)?'f':'o';
		return ret;
	}
	security("Invalid problem number!");
	return "crap."; /* <- never executed. */
}


int main(int argc,char *argv[])
{
	unsigned int port=80; /* use 0 to have OS pick a port */
	if (argv[1]) port=atoi(argv[1]);
	SERVER_SOCKET serv=skt_server(&port);
	setuid(1234); /* drop privilege */
	printf("Waiting for connect on port %d\n",port);
	
  while (1) {
    skt_select1(serv,0);
    printf("Incoming connection activity detected.\n"); fflush(stdout);
    s=skt_accept(serv,&client_ip,&client_port);
    int child=fork();
    if (child<0) {printf("forward: cannot spawn child\n"); exit(1);}
    if (child==0) { /* a connection is coming in--spawn a subprocess */
	skt_print_ip(client_name,client_ip);
	/** Log source IP, port, username, and request */
	printf("Connection from %s:%d\n",client_name,client_port); fflush(stdout);

// Print warning message
	std::string warning="Welcome to the CS 493/693 HW3 target.  Send your UAF email ID and a newline\n";
	skt_sendN(s,&warning[0],warning.size());

// Grab user name:	
	user=skt_recv_line(s); /* User name */
	printf("	User:'%s'\n",user.c_str()); fflush(stdout);
	if (user.size()>8u) security("user name too long!"); /* fsabc12 */
	if (user.size()<1) security("User name is empty."); 
	if (user[0]!='f') security("invalid user name!");

// Grab problem number
	std::string welcome="Which problem would you like to solve (1-5)?\n";
	skt_sendN(s,&welcome[0],welcome.size());
	std::string prob=skt_recv_line(s);
	if (prob.size()>1) security("Problem number too long!");
	int probNo=prob[0]-'0';
	if (probNo<1 || probNo>=5) security("Problem number out of range!");
	
// Pick a password and compute the hint
	printf("	Generating password for problem %d ('%s')\n",probNo,prob.c_str()); fflush(stdout);
	std::string password=generate_password(user,probNo);
	osl::SHA1_hash_t pw_hash=SHA1_hash(password.c_str());
	printf("	Generated password '%s' (%d digits) for problem %d\n",
		password.c_str(),password.size(),probNo);
	std::string hint=probInfo[probNo].hint+(std::string)".\n"
		"Password's SHA1 hash: "+SHA1_hex_from_hash(pw_hash)+"\n"
		"Please enter the password:\n";
	skt_sendN(s,&hint[0],hint.size());
	

	bool OK=false;
	//while (!OK) // Keep on tryin?
	{
// Retrieve the user's proported password
	std::string attempt=skt_recv_line(s);
	if (attempt.size()>16) security("Password attempt way too long."); 
	if (attempt.size()<1) security("Password attempt is empty."); 
	osl::SHA1_hash_t at_hash=SHA1_hash(attempt.c_str());
	const char *resp="NO--Sorry, that's not it.  Care to try again?\n";
	if (pw_hash==at_hash) {
		OK=true;
		resp="YES!  That's the password!  Problem complete.  Closing connection.\n";
	}
	
// Log result
	printf("%sProblem%d	User:'%s'	Attempt:'%s'	from:'%s:%d'\n",
		OK?"OK":"bad",probNo,user.c_str(),attempt.c_str(),
		client_name,client_port);fflush(stdout);
	
// Send result back to user
	skt_sendN(s,&resp[0],strlen(resp));
	}
	
	exit(0);
    } /* else parent-- wait for child? */
  }
}
